@@ -1,8 +1,9 @@
 ;;; emms-player-mpd.el --- MusicPD support for EMMS
 
-;; Copyright (C) 2005, 2006, 2007 Free Software Foundation, Inc.
+;; Copyright (C) 2005, 2006, 2007, 2008, 2009, 2014 Free Software Foundation, Inc.
 
-;; Author: Michael Olson <mwolson@gnu.org>
+;; Author: Michael Olson <mwolson@gnu.org>, Jose Antonio Ortega Ruiz
+;; <jao@gnu.org>
 
 ;; This file is part of EMMS.
 
@@ -102,9 +103,12 @@
 
 ;; Adam Sjøgren implemented support for changing the volume.
 
+(require 'cl-lib)
 (require 'emms-player-simple)
 (require 'emms-source-playlist)  ; for emms-source-file-parse-playlist
 (require 'tq)
+(require 'emms-cache)
+(require 'emms-url)
 
 (eval-when-compile
   (condition-case nil
@@ -128,8 +132,9 @@
 (defcustom emms-player-mpd-music-directory nil
   "The value of 'music_directory' in your MusicPD configuration file.
 
-You need this if your playlists use absolute file names, otherwise
-leave it set to nil."
+Unless your MusicPD is configured to use absolute file names, you must
+set this variable to the value of 'music_directory' in your MusicPD
+config."
   ;; The :format part ensures that entering directories happens on the
   ;; next line, where there is more space to work with
   :type '(choice :format "%{%t%}:\n   %[Value Menu%] %v"
@@ -140,31 +145,52 @@ leave it set to nil."
 (defun emms-player-mpd-get-supported-regexp ()
   "Returns a regexp of file extensions that MusicPD supports,
 or nil if we cannot figure it out."
-  (let ((out (split-string (shell-command-to-string "mpd --version")
-                           "\n"))
-        (found-start nil)
-        (supported nil))
-    ;; Get supported formats
-    (while (car out)
-      (cond ((string= (car out) "Supported formats:")
-             (setq found-start t))
-            ((string= (car out) "")
-             (setq found-start nil))
-            (found-start
-             (setq supported (concat supported (car out)))))
-      (setq out (cdr out)))
-    ;; Create regexp
-    (when (and (stringp supported)
-               (not (string= supported "")))
-      (concat "\\`http://\\|\\.\\(m3u\\|pls\\|"
-              (regexp-opt (delq nil (split-string supported)))
-              "\\)\\'"))))
+  (let ((out (shell-command-to-string "mpd --version")))
+    ;; 0.17.x
+    (if (string-match "Decoders plugins:$" out)
+        (let* ((b (match-end 0))
+               (e (string-match "Output plugins:$" out))
+               (plugs (split-string (substring out b e) "\n" t))
+               (plugs (cl-mapcan (lambda (x)
+                                   (and (string-match " +\\[.*\\] +\\(.+\\)$" x)
+                                        (split-string (match-string 1 x) nil t)))
+                                 plugs))
+               (b (and (string-match "Protocols:$" out) (match-end 0)))
+               (prots (and b (substring out (+ 2 b) -1)))
+               (prots (split-string (or prots "") nil t)))
+          (concat "\\(\\.\\(m3u\\|pls\\|"
+                  (regexp-opt (delq nil plugs))
+                  "\\)\\'\\)\\|\\(\\`"
+                  (regexp-opt (delete "file://" prots)) "\\)"))
+      (let ((found-start nil)
+            (supported nil))
+        (if (string-match "Supported decoders:\\([^0]+?\\)Supported outputs:" out)
+            ;; 0.15.x
+            (setq supported (replace-regexp-in-string "\\[.+?\\]" ""
+                                                      (match-string 1 out)))
+          ;; < 0.15
+          (setq out (split-string out "\n"))
+          (while (car out)
+            (cond ((string= (car out) "Supported formats:")
+                   (setq found-start t))
+                  ((string= (car out) "")
+                   (setq found-start nil))
+                  (found-start
+                   (setq supported (concat supported (car out)))))
+            (setq out (cdr out))))
+        ;; Create regexp
+        (when (and (stringp supported)
+                   (not (string= supported "")))
+          (concat "\\`http://\\|\\.\\(m3u\\|pls\\|"
+                  (regexp-opt (delq nil (split-string supported)))
+                  "\\)\\'"))))))
 
 (defcustom emms-player-mpd-supported-regexp
   ;; Use a sane default, just in case
   (or (emms-player-mpd-get-supported-regexp)
       (concat "\\`http://\\|"
-              "\\.\\(m3u\\|ogg\\|flac\\|mp3\\|wav\\|mod\\|au\\|aiff\\)\\'"))
+              (emms-player-simple-regexp
+               "m3u" "ogg" "flac" "mp3" "wav" "mod" "au" "aiff")))
   "Formats supported by MusicPD."
   :type 'regexp
   :set (function
@@ -197,10 +223,14 @@ It should take same arguments as `open-network-stream' does."
 
 (defcustom emms-player-mpd-check-interval 1
   "How often to check to see whether MusicPD has advanced to the
-next song.  This may be an integer or a floating point number.
+next song.  This may be an integer, a floating point number, or
+nil.  If set to nil, this check will not be periodically
+performed.
 
-This is used only if `emms-player-mpd-sync-playlist' is non-nil"
-  :type 'number
+This variable is used only if `emms-player-mpd-sync-playlist' is
+non-nil."
+  :type '(choice (const :tag "Disable check" nil)
+                 number)
   :group 'emms-player-mpd)
 
 (defcustom emms-player-mpd-verbose nil
@@ -248,6 +278,7 @@ If your EMMS playlist contains stored playlists, set this to nil."
 (make-variable-buffer-local 'emms-player-mpd-playlist-id)
 
 (defvar emms-player-mpd-current-song nil)
+(defvar emms-player-mpd-last-state nil)
 (defvar emms-player-mpd-status-timer nil)
 
 (defvar emms-player-mpd-status-regexp
@@ -272,27 +303,27 @@ return at the end of a request.")
            (when emms-player-mpd-verbose
              (message "Other MusicPD status change: %s, %s" status event))))))
 
-;; Ignore a useless byte-compile warning
-(eval-when-compile
-  (put 'process-kill-without-query 'byte-compile nil))
-
 (defun emms-player-mpd-ensure-process ()
   "Make sure that a MusicPD process is currently active."
   (unless (and emms-player-mpd-process
                (processp emms-player-mpd-process)
                (memq (process-status emms-player-mpd-process) '(run open)))
     (setq emms-player-mpd-process
-          (funcall emms-player-mpd-connect-function "mpd"
-                   nil
-                   emms-player-mpd-server-name
-                   emms-player-mpd-server-port))
+	  (if emms-player-mpd-server-port
+	      (funcall emms-player-mpd-connect-function "mpd"
+		     nil
+		     emms-player-mpd-server-name
+		     emms-player-mpd-server-port)
+	    (make-network-process :name "emms-mpd"
+				:service emms-player-mpd-server-name
+				:family 'local)))
     (set-process-sentinel emms-player-mpd-process
                           'emms-player-mpd-sentinel)
     (setq emms-player-mpd-queue
           (tq-create emms-player-mpd-process))
     (if (fboundp 'set-process-query-on-exit-flag)
         (set-process-query-on-exit-flag emms-player-mpd-process nil)
-      (process-kill-without-query emms-player-mpd-process))
+      (set-process-query-on-exit-flag emms-player-mpd-process nil))
     ;; send password
     (when (stringp emms-player-mpd-server-password)
       (tq-enqueue emms-player-mpd-queue
@@ -429,6 +460,7 @@ The list will be in reverse order."
                            'file))
                    (track (emms-track type file)))
               (emms-info-mpd track song-info)
+              (run-hook-with-args 'emms-track-info-filters track)
               (setq tracks (cons track tracks)))))))
     (funcall (car closure) (cdr closure) tracks)))
 
@@ -578,9 +610,10 @@ buffer and move to NEW-SONG."
    (lambda (closure id)
      (let ((buffer (car closure))
            (fn (cdr closure)))
-       (funcall fn buffer id)))))
+       (when (functionp fn)
+         (funcall fn buffer id))))))
 
-(defun emms-player-mpd-sync-from-emms (callback)
+(defun emms-player-mpd-sync-from-emms (&optional callback)
   "Synchronize the MusicPD playlist with the contents of the
 current EMMS playlist.
 
@@ -629,15 +662,30 @@ errors."
 
 (defun emms-player-mpd-sync-from-mpd (&optional closure callback)
   "Synchronize the EMMS playlist with the contents of the current
-MusicPD playlist."
+MusicPD playlist.  Namely, clear the EMMS playlist buffer and add
+tracks to it that are present in the MusicPD playlist.
+
+If the current buffer is an EMMS playlist buffer, make it the
+main EMMS playlist buffer."
+  (when (and emms-playlist-buffer-p
+             (not (eq (current-buffer) emms-playlist-buffer)))
+    (emms-playlist-set-playlist-buffer (current-buffer)))
   (with-current-emms-playlist
     (emms-player-mpd-get-tracks
      (cons emms-playlist-buffer (cons callback closure))
      #'emms-player-mpd-sync-from-mpd-1)))
 
+(defun emms-player-mpd-detect-song-change-2 (state info)
+  "Perform post-sync tasks after returning from a stop."
+  (setq emms-player-mpd-current-song nil)
+  (setq emms-player-playing-p 'emms-player-mpd)
+  (when (string= state "pause")
+    (setq emms-player-paused-p t))
+  (emms-player-mpd-detect-song-change info))
+
 (defun emms-player-mpd-detect-song-change-1 (closure info)
   (let ((song (emms-player-mpd-get-current-song nil #'ignore info))
-        (status (emms-player-mpd-get-mpd-state nil #'ignore info))
+        (state (emms-player-mpd-get-mpd-state nil #'ignore info))
         (time (emms-player-mpd-get-playing-time nil #'ignore info))
         (err-msg (cdr (assoc "error" info))))
     (if (stringp err-msg)
@@ -646,19 +694,29 @@ MusicPD playlist."
           (emms-player-mpd-send
            "clearerror"
            nil #'ignore))
-      (cond ((string= status "stop")
-             (emms-player-mpd-disconnect t)
+      (cond ((string= state "stop")
              (if song
                  ;; a track remains: the user probably stopped MusicPD
                  ;; manually, so we'll stop EMMS completely
                  (let ((emms-player-stopped-p t))
+                   (setq emms-player-mpd-last-state "stop")
                    (emms-player-stopped))
                ;; no more tracks are left: we probably ran out of things
                ;; to play, so let EMMS do something further if it wants
-               (emms-player-stopped)))
-            ((string= status "pause")
+               (unless (string= emms-player-mpd-last-state "stop")
+                 (setq emms-player-mpd-last-state "stop")
+                 (emms-player-stopped))))
+            ((and emms-player-mpd-last-state
+                  (string= emms-player-mpd-last-state "stop"))
+             ;; resume from a stop that occurred outside of EMMS
+             (setq emms-player-mpd-last-state nil)
+             (emms-player-mpd-sync-from-mpd
+              state
+              #'emms-player-mpd-detect-song-change-2))
+            ((string= state "pause")
              nil)
-            ((string= status "play")
+            ((string= state "play")
+             (setq emms-player-mpd-last-state "play")
              (unless (or (null song)
                          (and (stringp emms-player-mpd-current-song)
                               (string= song emms-player-mpd-current-song)))
@@ -689,12 +747,17 @@ info from MusicPD."
            (emms-replace-regexp-in-string "\\\\" "\\\\\\\\" file))
           "\""))
 
+;;;###autoload
 (defun emms-player-mpd-clear ()
-  "Clear the playlist."
+  "Clear the MusicPD playlist."
+  (interactive)
   (when emms-player-mpd-status-timer
     (emms-cancel-timer emms-player-mpd-status-timer)
     (setq emms-player-mpd-status-timer nil))
-  (emms-player-mpd-send "clear" nil #'ignore))
+  (setq emms-player-mpd-last-state nil)
+  (emms-player-mpd-send "clear" nil #'ignore)
+  (let ((emms-player-stopped-p t))
+    (emms-player-stopped)))
 
 ;;; Adding to the MusicPD playlist
 
@@ -737,7 +800,7 @@ Execute CALLBACK with CLOSURE as its first argument when done.
 This handles both m3u and pls type playlists."
   ;; This is useful for playlists of playlists
   (with-temp-buffer
-    (insert-file-contents playlist)
+    (emms-insert-file-contents playlist)
     (emms-player-mpd-add-buffer-contents (current-buffer) closure callback)))
 
 (defun emms-player-mpd-add-streamlist (url closure callback)
@@ -748,7 +811,7 @@ Execute CALLBACK with CLOSURE as its first argument when done."
       (progn
         (require 'emms-url)
         (with-temp-buffer
-          (url-insert-file-contents (emms-escape-url url))
+          (url-insert-file-contents (emms-url-quote-entire url))
           (emms-http-decode-buffer (current-buffer))
           (emms-player-mpd-add-buffer-contents (current-buffer)
                                                closure callback)))
@@ -767,7 +830,8 @@ Execute CALLBACK with CLOSURE as its first argument when done."
           ((or (eq type 'playlist)
                (string-match "\\.\\(m3u\\|pls\\)\\'" name))
            (emms-player-mpd-add-playlist name closure callback))
-          ((eq type 'file)
+          ((and (eq type 'file)
+                (string-match emms-player-mpd-supported-regexp name))
            (emms-player-mpd-add-file name closure callback)))))
 
 (defun emms-player-mpd-add-several-tracks (tracks closure callback)
@@ -806,6 +870,7 @@ Execute CALLBACK with CLOSURE as its first argument when done."
   "Play whatever is in the current MusicPD playlist.
 If ID is specified, play the song at that position in the MusicPD
 playlist."
+  (interactive)
   (if id
       (progn
         (unless (stringp id)
@@ -815,9 +880,11 @@ playlist."
          nil
          (lambda (closure response)
            (setq emms-player-mpd-current-song nil)
-           (setq emms-player-mpd-status-timer
-                 (run-at-time t emms-player-mpd-check-interval
-                              'emms-player-mpd-detect-song-change)))))
+           (if emms-player-mpd-check-interval
+               (setq emms-player-mpd-status-timer
+                     (run-at-time t emms-player-mpd-check-interval
+                                  'emms-player-mpd-detect-song-change))
+             (emms-player-mpd-detect-song-change)))))
     ;; we only want to play one track, so don't start the timer
     (emms-player-mpd-send
      "play"
@@ -880,9 +947,10 @@ This is called if `emms-player-mpd-sync-playlist' is non-nil."
       (setq emms-player-paused-p t))
     (unless (string= state "stop")
       (emms-player-mpd-detect-song-change info)
-      (setq emms-player-mpd-status-timer
-            (run-at-time t emms-player-mpd-check-interval
-                         'emms-player-mpd-detect-song-change)))))
+      (when emms-player-mpd-check-interval
+        (setq emms-player-mpd-status-timer
+              (run-at-time t emms-player-mpd-check-interval
+                           'emms-player-mpd-detect-song-change))))))
 
 ;;;###autoload
 (defun emms-player-mpd-connect ()
@@ -905,8 +973,7 @@ buffer get out-of-sync for some reason."
   "Starts a process playing TRACK."
   (interactive)
   (if (and emms-player-mpd-sync-playlist
-           (not (memq (emms-track-get track 'type) '(streamlist playlist)))
-           (not (string-match "\\`http://" (emms-track-get track 'name))))
+           (not (memq (emms-track-get track 'type) '(streamlist playlist))))
       (emms-player-mpd-start-and-sync)
     (emms-player-mpd-clear)
     ;; if we have loaded the item successfully, play it
@@ -922,6 +989,7 @@ from other functions."
   (emms-cancel-timer emms-player-mpd-status-timer)
   (setq emms-player-mpd-status-timer nil)
   (setq emms-player-mpd-current-song nil)
+  (setq emms-player-mpd-last-state nil)
   (emms-player-mpd-close-process)
   (unless no-stop
     (let ((emms-player-stopped-p t))
@@ -952,7 +1020,7 @@ from other functions."
            (secs (emms-player-mpd-get-playing-time nil #'ignore info)))
        (when (and song secs)
          (emms-player-mpd-send
-          (concat "seek " song " " (number-to-string (+ secs amount)))
+          (concat "seek " song " " (number-to-string (round (+ secs amount))))
           nil #'ignore))))))
 
 (defun emms-player-mpd-seek-to (pos)
@@ -963,7 +1031,7 @@ from other functions."
    (lambda (pos song)
      (when (and song pos)
        (emms-player-mpd-send
-        (concat "seek " song " " (number-to-string pos))
+        (concat "seek " song " " (number-to-string (round pos)))
         nil #'ignore)))))
 
 (defun emms-player-mpd-next ()
@@ -1020,6 +1088,7 @@ positive or negative."
               (emms-track-set track 'type 'file))
             (emms-track-set track 'name file)
             (emms-info-mpd track info)
+            (run-hook-with-args 'emms-track-info-filters track)
             (setq track-desc (emms-track-description track))
             (when (and (stringp track-desc) (not (string= track-desc "")))
               (setq desc (if desc
@@ -1064,9 +1133,12 @@ rather than EMMS."
     (let ((name (car data))
           (value (cdr data)))
       (setq name (cond ((string= name "artist") 'info-artist)
+                       ((string= name "composer") 'info-composer)
+                       ((string= name "performer") 'info-performer)
                        ((string= name "title") 'info-title)
                        ((string= name "album") 'info-album)
                        ((string= name "track") 'info-tracknumber)
+                       ((string= name "disc") 'info-discnumber)
                        ((string= name "date") 'info-year)
                        ((string= name "genre") 'info-genre)
                        ((string= name "time")
@@ -1092,7 +1164,7 @@ This is a useful addition to `emms-info-functions'."
   (if info
       (emms-info-mpd-process track info)
     (when (and (eq 'file (emms-track-type track))
-               (not (string-match "\\`http://" file)))
+               (not (string-match "\\`http://" (emms-track-name track))))
       (let ((file (emms-player-mpd-get-mpd-filename (emms-track-name track))))
         (when (or emms-player-mpd-music-directory
                   (and file
@@ -1121,6 +1193,14 @@ The track should be an alist as per `emms-player-mpd-get-alist'."
         (emms-info-mpd-process track track-info)
         (funcall emms-cache-set-function 'file name track)))))
 
+(defun emms-cache--info-cleanup (info)
+  (let ((xs (mapcar (lambda (x)
+                      (and (stringp x)
+                           (not (string-match-p "\\`\\(Last-\\|direct\\)" x))
+                           x))
+                    info)))
+    (cons nil (delq nil xs))))
+
 (defun emms-cache-set-from-mpd-directory (dir)
   "Dump all MusicPD data from DIR into the EMMS cache.
 
@@ -1140,11 +1220,18 @@ This is useful to do when you have recently acquired new music."
          nil
          (lambda (closure response)
            (message "Dumping MusicPD data to cache...processing")
-           (let ((info (emms-player-mpd-get-alists
-                        (emms-player-mpd-parse-response response))))
-             (dolist (track-info info)
-               (emms-cache-set-from-mpd-track track-info))
-             (message "Dumping MusicPD data to cache...done")))))
+           (let ((info (emms-player-mpd-parse-response response)))
+             (when (null (car info))
+               (let* ((info (emms-cache--info-cleanup info))
+                      (info (emms-player-mpd-get-alists info))
+                      (track 1)
+                      (total (length info)))
+                 (dolist (track-info info)
+                   (message "Dumping MusicPD data to cache...%d/%d" track total)
+                   (emms-cache-set-from-mpd-track track-info)
+                   (setq track (+ 1 track)))
+                 (message "Dumping MusicPD data to cache... %d tracks processed"
+                          total)))))))
     (error "Caching is not enabled")))
 
 (defun emms-cache-set-from-mpd-all ()
@@ -1167,7 +1254,7 @@ order to prime the cache."
   (unless (string= dir "")
     (setq dir (emms-player-mpd-get-mpd-filename dir)))
   (emms-player-mpd-send
-   (concat "update " dir) nil
+   (concat "update " (emms-player-mpd-quote-file dir)) nil
    (lambda (closure response)
      (let ((id (cdr (assoc "updating_db"
                            (emms-player-mpd-get-alist
@@ -1182,6 +1269,52 @@ the MusicPD database."
   (interactive)
   (emms-player-mpd-update-directory ""))
 
+(defvar emms-player-mpd-waiting-for-update-timer nil
+  "Timer object when waiting for MPD update to finish.")
+
+(defun emms-player-mpd-update-all-reset-cache ()
+  "Update all tracks in the MusicPD music directory.
+When update finishes, clear the EMMS cache and call
+`emms-cache-set-from-mpd-all' to dump the MusicPD data into the
+cache."
+  (interactive)
+  (if emms-player-mpd-waiting-for-update-timer
+      (message "Already waiting for an update to finish.")
+    (emms-player-mpd-send
+     "update" nil
+     'emms-player-mpd-wait-for-update)))
+
+(defun emms-player-mpd-wait-for-update (&optional closure response)
+  "Wait for a currently running mpd update to finish.
+Afterwards, clear the EMMS cache and call
+`emms-cache-set-from-mpd-all'."
+  (if response
+      ;; This is the first call after the update command
+      (let ((id (cdr (assoc "updating_db"
+			    (emms-player-mpd-get-alist
+			     (emms-player-mpd-parse-response response))))))
+	(if id
+	    (progn
+	      (message "Updating DB with ID %s.  Waiting for the update to finish..." id)
+	      (setq emms-player-mpd-waiting-for-update-timer
+		    (run-at-time 1 nil 'emms-player-mpd-wait-for-update)))
+	  (message "Could not update the DB")))
+    ;; Otherwise, check if update is still in progress
+    (emms-player-mpd-get-status-part
+     nil
+     (lambda (closure updating)
+       (if updating
+	   ;; MPD update still in progress, so wait another second
+	   (run-at-time 1 nil 'emms-player-mpd-wait-for-update)
+	 ;; MPD update finished
+	 (setq  emms-player-mpd-waiting-for-update-timer nil)
+	 (message "MPD update finished.")
+	 (sit-for 1)
+	 (clrhash emms-cache-db)
+	 (emms-cache-set-from-mpd-all)))
+     "updating_db")))
+
+
 (provide 'emms-player-mpd)
 
 ;;; emms-player-mpd.el ends here
