@@ -1,8 +1,8 @@
 ;;; emms-playlist-mode.el --- Playlist mode for Emms.
 
-;; Copyright (C) 2005, 2006, 2007 Free Software Foundation, Inc.
+;; Copyright (C) 2005-2019  Free Software Foundation, Inc.
 
-;; Author: Yoni Rabkin <yonirabkin@member.fsf.org>
+;; Author: Yoni Rabkin <yrk@gnu.org>
 
 ;; This file is part of EMMS.
 
@@ -35,10 +35,9 @@
 ;;; --------------------------------------------------------
 
 (require 'emms)
-(eval-when-compile
-  (condition-case nil
-      (require 'overlay)
-    (error nil)))
+(condition-case nil
+    (require 'overlay)
+  (error nil))
 (require 'emms-source-playlist)
 
 (defvar emms-playlist-mode-hook nil
@@ -53,6 +52,9 @@
 (defvar emms-playlist-mode-popup-enabled nil
   "True when the playlist was called as a popup window.")
 
+(defvar emms-playlist-mode-kill-whole-line-p t
+  "When true line kills behave like a typical music player.")
+
 (make-variable-buffer-local
  'emms-playlist-mode-selected-overlay)
 
@@ -110,7 +112,7 @@ This is true for every invocation of `emms-playlist-mode-go'."
 ;;; Keys
 ;;; --------------------------------------------------------
 
-(defconst emms-playlist-mode-map
+(defvar emms-playlist-mode-map
   (let ((map (make-sparse-keymap)))
     (set-keymap-parent map text-mode-map)
     (define-key map (kbd "C-x C-s") 'emms-playlist-save)
@@ -119,6 +121,7 @@ This is true for every invocation of `emms-playlist-mode-go'."
     (define-key map (kbd "C-w") 'emms-playlist-mode-kill)
     (define-key map (kbd "C-_") 'emms-playlist-mode-undo)
     (define-key map (kbd "C-/") 'emms-playlist-mode-undo)
+    (define-key map (kbd "C-x u") 'emms-playlist-mode-undo)
     (define-key map (kbd "C-n") 'next-line)
     (define-key map (kbd "C-p") 'previous-line)
     (define-key map (kbd "C-j") 'emms-playlist-mode-insert-newline)
@@ -129,7 +132,7 @@ This is true for every invocation of `emms-playlist-mode-go'."
     (define-key map (kbd "M-p") 'emms-playlist-mode-previous)
     (define-key map (kbd "a") 'emms-playlist-mode-add-contents)
     (define-key map (kbd "b") 'emms-playlist-set-playlist-buffer)
-    (define-key map (kbd "d") 'emms-playlist-mode-kill-entire-track)
+    (define-key map (kbd "D") 'emms-playlist-mode-kill-track)
     (define-key map (kbd "n") 'emms-next)
     (define-key map (kbd "p") 'emms-previous)
     (define-key map (kbd "SPC") 'scroll-up)
@@ -140,10 +143,11 @@ This is true for every invocation of `emms-playlist-mode-go'."
     (define-key map (kbd "f") 'emms-show)
     (define-key map (kbd "c") 'emms-playlist-mode-center-current)
     (define-key map (kbd "q") 'emms-playlist-mode-bury-buffer)
-    (define-key map (kbd "k") 'emms-playlist-current-kill)
+    (define-key map (kbd "K") 'emms-playlist-mode-current-kill)
     (define-key map (kbd "?") 'describe-mode)
     (define-key map (kbd "r") 'emms-random)
     (define-key map (kbd "C") 'emms-playlist-clear)
+    (define-key map (kbd "d") 'emms-playlist-mode-goto-dired-at-point)
     (define-key map (kbd "<mouse-2>") 'emms-playlist-mode-play-current-track)
     (define-key map (kbd "RET") 'emms-playlist-mode-play-smart)
     map)
@@ -178,6 +182,15 @@ FUN should be a function."
 	(setq emms-playlist-mode-popup-enabled nil))
     (bury-buffer)))
 
+(defun emms-playlist-mode-current-kill ()
+  "If the current buffer is an EMMS playlist buffer, kill it.
+Otherwise, kill the current EMMS playlist buffer."
+  (interactive)
+  (if (and emms-playlist-buffer-p
+           (not (eq (current-buffer) emms-playlist-buffer)))
+      (kill-buffer (current-buffer))
+    (emms-playlist-current-kill)))
+
 (defun emms-playlist-mode-last ()
   "Move to directly after the last track in the current buffer."
   (interactive)
@@ -195,15 +208,14 @@ FUN should be a function."
 (defun emms-playlist-mode-center-current ()
   "Move point to the currently selected track."
   (interactive)
-  (with-current-emms-playlist
-    (goto-char (if emms-playlist-mode-selected-overlay
-		   (overlay-start emms-playlist-mode-selected-overlay)
-		 (point-min)))))
+  (goto-char (if emms-playlist-mode-selected-overlay
+                 (overlay-start emms-playlist-mode-selected-overlay)
+               (point-min))))
 
 (defun emms-playlist-mode-play-current-track ()
   "Start playing track at point."
   (interactive)
-  (emms-playlist-set-playlist-buffer)
+  (emms-playlist-set-playlist-buffer (current-buffer))
   (unless (emms-playlist-track-at (point))
     (emms-playlist-next))
   (emms-playlist-select (point))
@@ -225,6 +237,8 @@ Otherwise play the track immediately."
     (emms-move-beginning-of-line nil)
     (if (not emms-playlist-mode-open-playlists)
         (emms-playlist-mode-play-current-track)
+      (unless (emms-playlist-track-at)
+        (emms-playlist-next))
       (let* ((track (emms-playlist-track-at))
              (name (emms-track-get track 'name))
              (type (emms-track-get track 'type)))
@@ -267,6 +281,8 @@ set it as current."
   (interactive)
   (save-excursion
     (emms-move-beginning-of-line nil)
+    (unless (emms-playlist-track-at)
+      (emms-playlist-next))
     (let* ((track (emms-playlist-track-at))
            (name (emms-track-get track 'name))
            (type (emms-track-get track 'type))
@@ -301,6 +317,18 @@ set it as current."
           (goto-char (point-min))
           (message "Added %s" (symbol-name type)))))))
 
+(defun emms-playlist-mode-goto-dired-at-point ()
+  "Visit the track at point in a `dired' buffer."
+  (interactive)
+  (let ((track (emms-playlist-track-at)))
+    (if track
+	(let ((name (emms-track-get track 'name))
+	      (type (emms-track-get track 'type)))
+	  (if (eq type 'file)
+	      (dired (file-name-directory name))
+	    (error "Can't visit this track type in Dired")))
+      (error "No track at point"))))
+
 ;;; --------------------------------------------------------
 ;;; Killing and yanking
 ;;; --------------------------------------------------------
@@ -310,7 +338,7 @@ set it as current."
   (and (<= a p)
        (<= p b)))
 
-;; d
+;; D
 (defun emms-playlist-mode-kill-entire-track ()
   "Kill track at point, including newline."
   (interactive)
@@ -318,23 +346,21 @@ set it as current."
     (emms-playlist-mode-kill-track)))
 
 ;; C-k
-;;
-;; Currently this kills as regular GNU/Emacs would and not like a
-;; typical music player would.
 (defun emms-playlist-mode-kill-track ()
   "Kill track at point."
   (interactive)
   (emms-with-inhibit-read-only-t
    (let ((track (emms-playlist-track-at)))
-     (if track
-	 (let ((track-region (emms-property-region (point)
-						   'emms-track)))
-	   (when (and emms-player-playing-p
-		      (emms-playlist-selected-track-at-p))
-	     (emms-stop)
-             (delete-overlay emms-playlist-mode-selected-overlay)
-             (setq emms-playlist-mode-selected-overlay nil))
-	   (kill-line))
+     (when track
+       (let ((track-region (emms-property-region (point)
+						 'emms-track)))
+	 (when (and emms-player-playing-p
+		    (emms-playlist-selected-track-at-p))
+	   (emms-stop)
+	   (delete-overlay emms-playlist-mode-selected-overlay)
+	   (setq emms-playlist-mode-selected-overlay nil))))
+     (let ((kill-whole-line emms-playlist-mode-kill-whole-line-p))
+       (goto-char (point-at-bol))
        (kill-line)))))
 
 ;; C-w
@@ -354,20 +380,28 @@ set it as current."
    (kill-region (region-beginning)
                 (region-end))))
 
+(defun emms-playlist-mode-correct-previous-yank ()
+  "Fix the previous yank if needed."
+  (when (and (< (point-at-bol) (point))
+	     (< (point) (point-at-eol)))
+    (newline)))
+
 ;; C-y
 (defun emms-playlist-mode-yank ()
   "Yank into the playlist buffer."
   (interactive)
   (emms-with-inhibit-read-only-t
    (goto-char (point-at-bol))
-   (yank)))
+   (yank)
+   (emms-playlist-mode-correct-previous-yank)))
 
 ;; M-y
 (defun emms-playlist-mode-yank-pop ()
   "Cycle through the kill-ring."
   (interactive)
   (emms-with-inhibit-read-only-t
-   (yank-pop nil)))
+   (yank-pop nil)
+   (emms-playlist-mode-correct-previous-yank)))
 
 ;;; --------------------------------------------------------
 ;;; Overlay
@@ -403,9 +437,10 @@ It creates a buffer called \"filename\", and restores the contents
 of the saved playlist inside."
   (interactive "fFile: ")
   (let* ((s)
-	 (buffer (find-file-noselect filename))
+	 (buffer (get-buffer-create filename))
 	 (name   (buffer-name buffer)))
     (with-current-buffer buffer
+      (emms-insert-file-contents filename)
       (setq s (read (buffer-string))))
     (kill-buffer buffer)
     (with-current-buffer (emms-playlist-new name)
@@ -491,7 +526,7 @@ playlist."
   (interactive)
   (if (or (null emms-playlist-buffer)
 	  (not (buffer-live-p emms-playlist-buffer)))
-      (call-interactively 'emms-add-file))
+      (call-interactively 'emms-add-directory))
   (emms-playlist-mode-go))
 
 (defun emms-playlist-mode-go-popup (&optional window-width)
@@ -501,7 +536,7 @@ Default value for WINDOW-WIDTH is `emms-playlist-mode-window-width'.
 WINDOW-WIDTH should be a positive integer."
   (interactive)
   (setq emms-playlist-mode-window-width
-	(or window-width emms-playlist-mode-window-width))
+        (round (or window-width emms-playlist-mode-window-width)))
   (split-window-horizontally (- emms-playlist-mode-window-width))
   (other-window 1)
   (emms-playlist-mode-go)
@@ -547,6 +582,7 @@ WINDOW-WIDTH should be a positive integer."
                         (point-max)
                         '(face emms-playlist-track-face)))
   (setq buffer-read-only t)
+  (setq truncate-lines t)
   (setq buffer-undo-list nil))
 
 ;;;###autoload
@@ -560,7 +596,7 @@ WINDOW-WIDTH should be a positive integer."
 
   (use-local-map emms-playlist-mode-map)
   (setq major-mode 'emms-playlist-mode
-	mode-name "Emms-Playlist")
+	mode-name "EMMS")
 
   (setq emms-playlist-insert-track-function
 	'emms-playlist-mode-insert-track)
