@@ -1,6 +1,7 @@
 ;;; emms-source-playlist.el --- EMMS sources from playlist files
 
-;; Copyright (C) 2003, 2004, 2005, 2006, 2007 Free Software Foundation, Inc.
+;; Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008,
+;;   2009 Free Software Foundation, Inc.
 
 ;; Author: Jorgen Schäfer <forcer@forcix.cx>
 ;; Keywords: emms, mp3, mpeg, multimedia
@@ -71,6 +72,16 @@ If this is nil, you will be prompted for a format to use."
                  (symbol :tag "Other"))
   :group 'emms)
 
+(defcustom emms-source-playlist-ask-before-overwrite t
+  "*Ask before saving over an existing playlist.
+If this is nil, existing playlists will be quietly overwritten."
+  :type 'boolean
+  :group 'emms)
+
+(defvar emms-source-playlist-native-header-line
+  ";;; This is an EMMS playlist file"
+  "Line which identifies a native emms playlist.")
+
 ;;; General playlist
 
 (defsubst emms-source-playlist-p-sym (format)
@@ -92,11 +103,11 @@ If this is nil, you will be prompted for a format to use."
                    (functionp (emms-source-playlist-files-sym format))))
       (funcall sym))))
 
-(defun emms-source-playlist-parse (format)
-  (funcall (emms-source-playlist-parse-sym format)))
+(defun emms-source-playlist-parse (format file)
+  (funcall (emms-source-playlist-parse-sym format) file))
 
-(defun emms-source-playlist-unparse (format playlist file)
-  (funcall (emms-source-playlist-unparse-sym format) playlist file))
+(defun emms-source-playlist-unparse (format playlist-buf file-buf)
+  (funcall (emms-source-playlist-unparse-sym format) playlist-buf file-buf))
 
 (defun emms-source-playlist-files (format)
   (let ((sym (emms-source-playlist-files-sym format)))
@@ -112,18 +123,25 @@ If this is nil, you will be prompted for a format to use."
 If `emms-source-playlist-default-format' is non-nil, use it
 instead of prompting the user."
   (or emms-source-playlist-default-format
-      (intern
-       (completing-read
-        (concat "Playlist format: (default: "
-                (if emms-source-playlist-format-history
-                    (car emms-source-playlist-format-history)
-                  "native")
-                ") ")
-        (mapcar #'symbol-name emms-source-playlist-formats)
-        nil nil nil 'emms-source-playlist-format-history
-        (if emms-source-playlist-format-history
-            (car emms-source-playlist-format-history)
-          "native")))))
+      (let ((format
+             (emms-completing-read
+              (concat "Playlist format: (default: "
+                      (if emms-source-playlist-format-history
+                          (car emms-source-playlist-format-history)
+                        "native")
+                      ") ")
+              (mapcar #'symbol-name emms-source-playlist-formats)
+              nil nil nil 'emms-source-playlist-format-history
+              (if emms-source-playlist-format-history
+                  (car emms-source-playlist-format-history)
+                "native"))))
+        ;; Sometimes the completion function can put partial results
+        ;; onto the history, so pop the last one off and include the
+        ;; completed version instead.
+        (setq emms-source-playlist-format-history
+              (cons format
+                    (cdr emms-source-playlist-format-history)))
+      (intern format))))
 
 (defun emms-playlist-save (format file)
   "Store the current playlist to FILE as the type FORMAT.
@@ -133,13 +151,19 @@ The default format is specified by `emms-source-playlist-default-format'."
                                      emms-source-file-default-directory
                                      emms-source-file-default-directory
                                      nil)))
-  (with-temp-buffer
-    (emms-source-playlist-unparse format
-                                  (with-current-emms-playlist
-                                    (current-buffer))
-                                  (current-buffer))
-    (let ((backup-inhibited t))
-      (write-file file))))
+  (if (or (eq emms-playlist-buffer (current-buffer))
+	  (and (not (eq emms-playlist-buffer (current-buffer)))
+	       (y-or-n-p
+		(format "Current playlist buffer (%s) is not the one you are visiting (%s). Save anyway?"
+			emms-playlist-buffer (current-buffer)))))
+      (with-temp-buffer
+	(emms-source-playlist-unparse format
+                                      (with-current-emms-playlist
+					(current-buffer))
+                                      (current-buffer))
+	(let ((backup-inhibited t))
+	  (write-file file emms-source-playlist-ask-before-overwrite)))
+    (message "aborting save")))
 
 (defun emms-source-playlist-determine-format (&optional parse-files)
   "Determine the playlist format of the current buffer.
@@ -163,26 +187,26 @@ See `emms-source-playlist-formats' for a list of supported formats."
                                      t)))
   (mapc #'emms-playlist-insert-track
         (with-temp-buffer
-          (insert-file-contents file)
+          (emms-insert-file-contents file)
           (goto-char (point-min))
           (let ((format (emms-source-playlist-determine-format)))
             (if format
-                (emms-source-playlist-parse format)
+                (emms-source-playlist-parse format file)
               (error "Not a recognized playlist format"))))))
 
-;;; EMMS native playlists
+;;; Emms native playlists
 
-;; Format:
-;; ;;; This is an EMMS playlist file. Play it with M-x emms-play-playlist
-;; <sexpr>
+;; An Emms native playlist file starts with the contents of
+;; `emms-source-playlist-native-header-line' and is followed by
+;; tracks in sexp format.
 
 (defun emms-source-playlist-native-p ()
   "Return non-nil if the current buffer contains a native EMMS playlist."
   (save-excursion
     (goto-char (point-min))
-    (looking-at "^;;; This is an EMMS playlist file")))
+    (looking-at (concat "^" emms-source-playlist-native-header-line))))
 
-(defun emms-source-playlist-parse-native ()
+(defun emms-source-playlist-parse-native (file)
   "Parse the native EMMS playlist in the current buffer."
   (save-excursion
     (goto-char (point-min))
@@ -195,17 +219,19 @@ OUT should be the buffer where tracks are stored in the native EMMS format."
   (with-current-buffer in ;; Don't modify the position
     (save-excursion       ;; in the IN buffer
       (with-current-buffer out
-        (insert ";;; This is an EMMS playlist file."
+        (insert emms-source-playlist-native-header-line
                 " Play it with M-x emms-play-playlist\n")
-        (insert "(")
-        (let ((track (emms-source-playlist-first in))
-              (firstp t))
-          (while track
-            (if (not firstp)
-                (insert "\n ")
-              (setq firstp nil))
-            (prin1 track (current-buffer))
-            (setq track (emms-source-playlist-next in))))
+        (insert "("))
+      (let ((firstp t))
+        (goto-char (point-min))
+        (emms-walk-tracks
+          (let ((track (emms-playlist-track-at (point))))
+            (with-current-buffer out
+              (if (not firstp)
+                  (insert "\n ")
+                (setq firstp nil))
+              (prin1 track (current-buffer))))))
+      (with-current-buffer out
         (insert ")\n")))))
 
 ;;;###autoload (autoload 'emms-play-native-playlist "emms-source-playlist" nil t)
@@ -218,11 +244,11 @@ OUT should be the buffer where tracks are stored in the native EMMS format."
                                      t)))
   (mapc #'emms-playlist-insert-track
         (with-temp-buffer
-          (insert-file-contents file)
+          (emms-insert-file-contents file)
           (goto-char (point-min))
           (when (not (emms-source-playlist-native-p))
             (error "Not a native EMMS playlist file."))
-          (emms-source-playlist-parse-native))))
+          (emms-source-playlist-parse-native file))))
 
 ;;; m3u files
 
@@ -247,13 +273,16 @@ the more restrictive playlist formats have already been
 detected and simply return non-nil always."
   t)
 
-(defun emms-source-playlist-parse-m3u ()
-  "Parse the m3u playlist in the current buffer."
-  (mapcar (lambda (file)
-            (if (string-match "\\`http://\\|\\`mms://" file)
-                (emms-track 'url file)
-              (emms-track 'file file)))
-          (emms-source-playlist-m3u-files)))
+(defun emms-source-playlist-parse-m3u (playlist-file)
+  "Parse the m3u playlist in the current buffer.
+Files will be relative to the directory of PLAYLIST-FILE, unless
+they have absolute paths."
+  (let ((dir (file-name-directory playlist-file)))
+    (mapcar (lambda (file)
+              (if (string-match "\\`\\(http[s]?\\|mms\\)://" file)
+                  (emms-track 'url file)
+                (emms-track 'file (expand-file-name file dir))))
+            (emms-source-playlist-m3u-files))))
 
 (defun emms-source-playlist-m3u-files ()
   "Extract a list of filenames from the given m3u playlist.
@@ -272,11 +301,11 @@ IN should be a buffer containing an m3u playlist.
 OUT should be the buffer where tracks are stored in m3u format."
   (with-current-buffer in ;; Don't modify the position
     (save-excursion       ;; in the IN buffer
-      (with-current-buffer out
-        (let ((track (emms-source-playlist-first in)))
-          (while track
-            (insert (emms-track-name track) ?\n)
-            (setq track (emms-source-playlist-next in))))))))
+      (goto-char (point-min))
+      (emms-walk-tracks
+        (let ((track (emms-playlist-track-at (point))))
+          (with-current-buffer out
+            (insert (emms-track-name track) ?\n)))))))
 
 ;;;###autoload (autoload 'emms-play-m3u-playlist "emms-source-playlist" nil t)
 ;;;###autoload (autoload 'emms-add-m3u-playlist "emms-source-playlist" nil t)
@@ -288,11 +317,11 @@ OUT should be the buffer where tracks are stored in m3u format."
                                      t)))
   (mapc #'emms-playlist-insert-track
         (with-temp-buffer
-          (insert-file-contents file)
+          (emms-insert-file-contents file)
           (goto-char (point-min))
           (when (not (emms-source-playlist-m3u-p))
             (error "Not an m3u playlist file."))
-          (emms-source-playlist-parse-m3u))))
+          (emms-source-playlist-parse-m3u file))))
 
 ;;; pls files
 
@@ -315,13 +344,20 @@ OUT should be the buffer where tracks are stored in m3u format."
         t
       nil)))
 
-(defun emms-source-playlist-parse-pls ()
-  "Parse the pls playlist in the current buffer."
-  (mapcar (lambda (file)
-            (if (string-match "\\`http://\\|\\`mms://" file)
-                (emms-track 'url file)
-              (emms-track 'file file)))
-          (emms-source-playlist-pls-files)))
+(defun emms-source-playlist-parse-pls (playlist-file)
+  "Parse the pls playlist in the current buffer.
+Files will be relative to the directory of PLAYLIST-FILE, unless
+they have absolute paths."
+  (let ((dir (file-name-directory playlist-file)))
+    (mapcar (lambda (file)
+              (if (string-match "\\`\\(http[s]?\\|mms\\)://" file)
+                  (emms-track 'url file)
+		(if (string-match "\\`file://" file) ;; handle file:// uris 
+		  (let ((file (url-unhex-string (substring file 7))))
+		    (emms-track 'file file))
+		  (emms-track 'file (expand-file-name file dir)))))
+      (emms-source-playlist-pls-files))))
+
 
 (defun emms-source-playlist-pls-files ()
   "Extract a list of filenames from the given pls playlist.
@@ -340,19 +376,21 @@ IN should be a buffer conatining a pls playlist.
 OUT should be the buffer where tracks are stored in pls format."
   (with-current-buffer in ;; Don't modify the position
     (save-excursion       ;; in the IN buffer
-      (with-current-buffer out
-        (let ((pos 0))
+      (let ((pos 0)
+            beg)
+        (with-current-buffer out
           (insert "[playlist]\n")
-          (save-restriction
-            (narrow-to-region (point) (point))
-            (let ((track (emms-source-playlist-first in)))
-              (while track
-                (setq pos (1+ pos))
-                (insert "File" (number-to-string pos) "="
-                        (emms-track-name track) ?\n)
-                (setq track (emms-source-playlist-next in))))
-            (goto-char (point-min))
-            (insert "NumberOfEntries=" (number-to-string pos) ?\n)))))))
+          (setq beg (point)))
+        (goto-char (point-min))
+        (emms-walk-tracks
+          (let ((track (emms-playlist-track-at (point))))
+            (setq pos (1+ pos))
+            (with-current-buffer out
+              (insert "File" (number-to-string pos) "="
+                      (emms-track-name track) ?\n))))
+        (with-current-buffer out
+          (goto-char beg)
+          (insert "NumberOfEntries=" (number-to-string pos) ?\n))))))
 
 ;;;###autoload (autoload 'emms-play-pls-playlist "emms-source-playlist" nil t)
 ;;;###autoload (autoload 'emms-add-pls-playlist "emms-source-playlist" nil t)
@@ -364,11 +402,11 @@ OUT should be the buffer where tracks are stored in pls format."
                                      t)))
   (mapc #'emms-playlist-insert-track
         (with-temp-buffer
-          (insert-file-contents file)
+          (emms-insert-file-contents file)
           (goto-char (point-min))
           (when (not (emms-source-playlist-pls-p))
             (error "Not a pls playlist file."))
-          (emms-source-playlist-parse-pls))))
+          (emms-source-playlist-parse-pls file))))
 
 ;;; extm3u files
 
@@ -396,30 +434,6 @@ OUT should be the buffer where tracks are stored in pls format."
 
 ;; Not implemented yet
 
-;;; Helper functions
-
-(defun emms-source-playlist-first (buf)
-  "Return the first track in BUF.
-This moves point."
-  (with-current-buffer buf
-    (condition-case nil
-        (progn
-          (emms-playlist-first)
-          (emms-playlist-track-at (point)))
-      (error
-       nil))))
-
-(defun emms-source-playlist-next (buf)
-  "Return the next track in BUF.
-This moves point."
-  (with-current-buffer buf
-    (condition-case nil
-        (progn
-          (emms-playlist-next)
-          (emms-playlist-track-at (point)))
-      (error
-       nil))))
-
 ;;; Adding playlists as files
 
 ;;;###autoload (autoload 'emms-play-playlist-file "emms-source-playlist" nil t)
