@@ -1,26 +1,30 @@
 ;;; emms.el --- The Emacs Multimedia System
 
-;; Copyright (C) 2003, 2004, 2005, 2006, 2007 Free Software Foundation, Inc.
+;; Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008,
+;;   2009, 2018, 2020  Free Software Foundation, Inc.
 
-;; Author: Jorgen Schfer <forcer@forcix.cx>
-;; Keywords: emms, mp3, mpeg, multimedia
+;; Author: Jorgen Schäfer <forcer@forcix.cx>, the Emms developers (see AUTHORS file)
+;; Maintainer: Yoni Rabkin <yrk@gnu.org>
+;; Version: 6.00
+;; Keywords: emms, mp3, ogg, flac, music, mpeg, video, multimedia
+;; Package-Type: multi
+;; Package-Requires: ((cl-lib "0.5"))
+;; url: https://www.gnu.org/software/emms/
 
 ;; This file is part of EMMS.
 
-;; EMMS is free software; you can redistribute it and/or modify
-;; it under the terms of the GNU General Public License as published by
+;; EMMS is free software; you can redistribute it and/or modify it
+;; under the terms of the GNU General Public License as published by
 ;; the Free Software Foundation; either version 3, or (at your option)
 ;; any later version.
-;;
-;; EMMS is distributed in the hope that it will be useful,
-;; but WITHOUT ANY WARRANTY; without even the implied warranty of
-;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-;; GNU General Public License for more details.
-;;
+
+;; EMMS is distributed in the hope that it will be useful, but WITHOUT
+;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
+;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
+;; License for more details.
+
 ;; You should have received a copy of the GNU General Public License
-;; along with EMMS; see the file COPYING.  If not, write to the
-;; Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
-;; Boston, MA 02110-1301, USA.
+;; along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
 
@@ -34,15 +38,17 @@
 ;; even know how to find any tracks to consider playing --- for this,
 ;; you need sources.
 
-;; A sample configuration is offered in emms-setup.el, so you might
-;; just want to use that file.
+;; A sample configuration is offered in emms-setup.el, and the
+;; Friendly Manual in the doc/ directory is both detailed, and kept up
+;; to date.
+
 
 ;;; Code:
 
-(defvar emms-version "3.0"
+(defvar emms-version "6.00"
   "EMMS version string.")
 
-
+
 ;;; User Customization
 
 (defgroup emms nil
@@ -80,12 +86,34 @@ If non-nil, EMMS will wrap back to the first track when that happens."
   :group 'emms
   :type 'boolean)
 
+(defcustom emms-random-playlist nil
+  "*Non-nil means that tracks are played randomly. If nil, tracks
+are played sequentially."
+  :group 'emms
+  :type 'boolean)
+
 (defcustom emms-repeat-track nil
   "Non-nil, playback will repeat current track.  If nil, EMMS will play
 track by track normally."
   :group 'emms
   :type 'boolean)
 
+(defvar-local emms-single-track nil
+  "Non-nil, play the current track and then stop.")
+
+(defcustom emms-completing-read-function
+  (if (and (boundp 'ido-mode)
+           ido-mode)
+      'ido-completing-read
+    'completing-read)
+  "Function to call when prompting user to choose between a list of options.
+This should take the same arguments as `completing-read'.  Some
+possible values are `completing-read' and `ido-completing-read'.
+Note that you must set `ido-mode' if using
+`ido-completing-read'."
+  :group 'emms
+  :type 'function)
+
 (defcustom emms-track-description-function 'emms-track-simple-description
   "*Function for describing an EMMS track in a user-friendly way."
   :group 'emms
@@ -93,10 +121,10 @@ track by track normally."
 
 (defcustom emms-player-delay 0
   "The delay to pause after a player finished.
-This is a floating-point number of seconds.
-This is necessary for some platforms where it takes a bit to free
-the audio device after a player has finished. If EMMS is skipping
-songs, increase this number."
+This is a floating-point number of seconds.  This is necessary
+for some platforms where it takes a bit to free the audio device
+after a player has finished.  If EMMS is skipping songs, increase
+this number."
   :type 'number
   :group 'emms)
 
@@ -111,7 +139,7 @@ songs, increase this number."
   :group 'emms)
 
 (defcustom emms-playlist-uniq-function 'emms-playlist-simple-uniq
-  "*The function to use for make track uniq in the playlist."
+  "*The function to use for removing duplicate tracks in the playlist."
   :type 'function
   :group 'emms)
 
@@ -127,7 +155,7 @@ sorts before the second (see `sort')."
   :type 'string
   :group 'emms)
 
-(defcustom emms-playlist-default-major-mode default-major-mode
+(defcustom emms-playlist-default-major-mode 'emms-playlist-mode
   "*The default major mode for EMMS playlist."
   :type 'function
   :group 'emms)
@@ -140,7 +168,7 @@ sorts before the second (see `sort')."
 
 (defcustom emms-playlist-update-track-function 'emms-playlist-simple-update-track
   "*A function to update the track at point.
-This is called when the track information changed. This also
+This is called when the track information changed.  This also
 shouldn't assume that the track has been inserted before."
   :group 'emms
   :type 'function)
@@ -152,6 +180,11 @@ shouldn't assume that the track has been inserted before."
   :type 'function)
 (make-variable-buffer-local 'emms-playlist-delete-track-function)
 
+(defcustom emms-ok-track-function 'emms-default-ok-track-function
+  "*Function returns true if we shouldn't skip this track."
+  :group 'emms
+  :type 'function)
+
 (defcustom emms-playlist-source-inserted-hook nil
   "*Hook run when a source got inserted into the playlist.
 The buffer is narrowed to the new tracks."
@@ -176,8 +209,16 @@ This can be used to initialize tracks with various info."
   :group 'emms
   :type 'hook)
 
+(defcustom emms-track-info-filters nil
+  "*List of functions to call when a track changes data, before updating
+the display.
+These functions are passed the track as an argument."
+  :group 'emms
+  :type 'hook)
+
 (defcustom emms-track-updated-functions nil
-  "*List of functions to call when a track changes data.
+  "*List of functions to call when a track changes data, after updating
+the display.
 These functions are passed the track as an argument."
   :group 'emms
   :type 'hook)
@@ -197,9 +238,9 @@ See `emms-player-finished-hook'."
 (defcustom emms-player-finished-hook nil
   "*Hook run when an EMMS player finishes playing a track.
 Please pay attention to the differences between
-`emms-player-finished-hook' and `emms-player-stopped-hook'.
-The former is called only when the player actually finishes
-playing a track; the latter, only when the player is stopped
+`emms-player-finished-hook' and `emms-player-stopped-hook'.  The
+former is called only when the player actually finishes playing a
+track; the latter, only when the player is stopped
 interactively."
   :group 'emms
   :type 'hook)
@@ -217,6 +258,11 @@ Use `emms-player-paused-p' to find the current state."
   :group 'emms
   :type 'hook)
 
+(defcustom emms-seek-seconds 10
+  "The number of seconds to seek forward or backward when seeking."
+  :group 'emms
+  :type 'number)
+
 (defcustom emms-player-seeked-functions nil
   "*Functions called when a player is seeking.
 The functions are called with a single argument, the amount of
@@ -250,6 +296,11 @@ The modified track is passed as the argument to this function."
   :group 'emms
   :type 'function)
 
+(defcustom emms-directory (expand-file-name "emms" user-emacs-directory)
+  "*Directory variable from which all other emms file variables are derived."
+  :group 'emms
+  :type 'string)
+
 (defvar emms-player-playing-p nil
   "The currently playing EMMS player, or nil.")
 
@@ -278,9 +329,9 @@ This also disables any read-onliness of the current buffer."
        (emms-playlist-current-clear))
      (let ((emms-source-old-buffer (or emms-source-old-buffer
                                        (current-buffer))))
-      (with-current-buffer emms-playlist-buffer
-        (let ((inhibit-read-only t))
-          ,@body)))))
+       (with-current-buffer emms-playlist-buffer
+	 (let ((inhibit-read-only t))
+	   ,@body)))))
 (put 'with-current-emms-playlist 'lisp-indent-function 0)
 (put 'with-current-emms-playlist 'edebug-form-spec '(body))
 
@@ -298,10 +349,10 @@ This also disables any read-onliness of the current buffer."
 
 (defmacro emms-walk-tracks (&rest body)
   "Execute BODY for each track in the current buffer, starting at point.
-The point will be placed at the beginning of the track before
+Point will be placed at the beginning of the track before
 executing BODY.
 
-The point will not be restored afterward."
+Point will not be restored afterward."
   (let ((donep (make-symbol "donep")))
     `(let ((,donep nil))
        ;; skip to first track if not on one
@@ -320,6 +371,13 @@ The point will not be restored afterward."
 (put 'emms-walk-tracks 'lisp-indent-function 0)
 (put 'emms-walk-tracks 'edebug-form-spec '(body))
 
+(defvar emms-player-base-format-list
+  '("ogg" "mp3" "wav" "mpg" "mpeg" "wmv" "wma"
+    "mov" "avi" "divx" "ogm" "ogv" "asf" "mkv"
+    "rm" "rmvb" "mp4" "flac" "vob" "m4a" "ape"
+    "flv" "webm" "aif")
+  "A list of common formats which player definitions can use.")
+
 
 ;;; User Interface
 
@@ -356,12 +414,18 @@ This is a good function to put in `emms-player-next-function'."
     (error "A track is already being played"))
   (cond (emms-repeat-track
 	 (emms-start))
+	(emms-single-track		; buffer local
+	 (emms-stop))
+	;; attempt to play the next track but ignore errors
 	((condition-case nil
              (progn
                (emms-playlist-current-select-next)
                t)
            (error nil))
-	 (emms-start))
+	 (if (funcall emms-ok-track-function
+		      (emms-playlist-current-selected-track))
+	     (emms-start)
+	   (emms-next-noerror)))
         (t
 	 (message "No next track in playlist"))))
 
@@ -382,10 +446,12 @@ This is a good function to put in `emms-player-next-function'."
   (emms-start))
 
 (defun emms-pause ()
-  "Pause the current player."
+  "Pause the current player.
+If player hasn't started, then start it now."
   (interactive)
-  (when emms-player-playing-p
-    (emms-player-pause)))
+  (if emms-player-playing-p
+      (emms-player-pause)
+    (emms-start)))
 
 (defun emms-seek (seconds)
   "Seek the current player SECONDS seconds.
@@ -407,13 +473,13 @@ It can also be negative to seek backwards."
   "Seek ten seconds forward."
   (interactive)
   (when emms-player-playing-p
-    (emms-player-seek 10)))
+    (emms-player-seek emms-seek-seconds)))
 
 (defun emms-seek-backward ()
   "Seek ten seconds backward."
   (interactive)
   (when emms-player-playing-p
-    (emms-player-seek -10)))
+    (emms-player-seek (- emms-seek-seconds))))
 
 (defun emms-show (&optional insertp)
   "Describe the current EMMS track in the minibuffer.
@@ -446,13 +512,36 @@ This uses `emms-playlist-sort-function'."
       (funcall emms-playlist-sort-function))))
 
 (defun emms-uniq ()
-  "Uniq the current playlist.
+  "Remove duplicates from the current playlist.
 This uses `emms-playlist-uniq-function'."
   (interactive)
   (with-current-emms-playlist
     (save-excursion
       (funcall emms-playlist-uniq-function))))
 
+(defun emms-toggle-single-track ()
+  "Toggle if Emms plays a single track and stops."
+  (interactive)
+  (with-current-emms-playlist
+    (cond (emms-single-track
+	   (setq emms-single-track nil)
+	   (message "single track mode disabled for %s"
+		    (buffer-name)))
+	  (t (setq emms-single-track t)
+	     (message "single track mode enabled for %s"
+		    (buffer-name))))))
+
+(defun emms-toggle-random-playlist ()
+  "Toggle whether emms plays the tracks randomly or sequentially.
+See `emms-random-playlist'."
+  (interactive)
+  (setq emms-random-playlist (not emms-random-playlist))
+  (if emms-random-playlist
+      (progn (setq emms-player-next-function 'emms-random)
+             (message "Will play the tracks randomly."))
+    (setq emms-player-next-function 'emms-next-noerror)
+    (message "Will play the tracks sequentially.")))
+
 (defun emms-toggle-repeat-playlist ()
   "Toggle whether emms repeats the playlist after it is done.
 See `emms-repeat-playlist'."
@@ -464,7 +553,7 @@ See `emms-repeat-playlist'."
 
 (defun emms-toggle-repeat-track ()
   "Toggle whether emms repeats the current track.
-See  `emms-repeat-track'."
+See `emms-repeat-track'."
   (interactive)
   (setq emms-repeat-track (not emms-repeat-track))
   (if emms-repeat-track
@@ -481,11 +570,50 @@ See  `emms-repeat-track'."
   (when (not emms-player-playing-p)
     (error "No EMMS player playing right now")))
 
+(defun emms-completing-read (&rest args)
+  "Read a string in the minibuffer, with completion.
+Set `emms-completing-read' to determine which function to use.
+
+See `completing-read' for a description of ARGS."
+  (apply emms-completing-read-function args))
+
+(defun emms-display-modes ()
+  "Display the current EMMS play modes."
+  (interactive)
+  (with-current-emms-playlist
+    (message
+     "repeat playlist: %s, repeat track: %s, random: %s, single %s"
+     (if emms-repeat-playlist "yes" "no")
+     (if emms-repeat-track "yes" "no")
+     (if emms-random-playlist "yes" "no")
+     (if emms-single-track "yes" "no"))))
+
 
 ;;; Compatibility functions
 
 (require 'emms-compat)
 
+
+;;; Utility functions
+
+(defun emms-insert-file-contents (filename &optional visit)
+  "Insert the contents of file FILENAME after point.
+Do character code conversion and end-of-line conversion, but none
+of the other unnecessary things like format decoding or
+`find-file-hook'.
+
+If VISIT is non-nil, the buffer's visited filename
+and last save file modtime are set, and it is marked unmodified.
+If visiting and the file does not exist, visiting is completed
+before the error is signaled."
+  (let ((format-alist nil)
+        (after-insert-file-functions nil)
+        (inhibit-file-name-handlers
+         (append '(jka-compr-handler image-file-handler epa-file-handler)
+                 inhibit-file-name-handlers))
+        (inhibit-file-name-operation 'insert-file-contents))
+    (insert-file-contents filename visit)))
+
 
 ;;; Dictionaries
 
@@ -566,8 +694,9 @@ If there is no value, return DEFAULT (or nil, if not given)."
 
 (defun emms-track-description (track)
   "Return a description of TRACK.
-This function uses the global value for `emms-track-description-function',
-rather than anything the current mode might have set.
+This function uses the global value for
+`emms-track-description-function', rather than anything the
+current mode might have set.
 
 Use `emms-track-force-description' instead if you need to insert
 a description into a playlist buffer."
@@ -575,17 +704,27 @@ a description into a playlist buffer."
 
 (defun emms-track-updated (track)
   "Information in TRACK got updated."
+  (run-hook-with-args 'emms-track-info-filters track)
   (emms-playlist-track-updated track)
   (run-hook-with-args 'emms-track-updated-functions track))
 
 (defun emms-track-simple-description (track)
   "Simple function to give a user-readable description of a track.
-If it's a file track, just return the file name.
-Otherwise, return the type and the name with a colon in between."
-  (if (eq 'file (emms-track-type track))
-      (emms-track-name track)
-    (concat (symbol-name (emms-track-type track))
-            ": " (emms-track-name track))))
+If it's a file track, just return the file name.  Otherwise,
+return the type and the name with a colon in between.
+Hex-encoded characters in URLs are replaced by the decoded
+character."
+  (let ((type (emms-track-type track)))
+    (cond ((eq 'file type)
+           (emms-track-name track))
+          ((eq 'url type)
+           (emms-format-url-track-name (emms-track-name track)))
+          (t (concat (symbol-name type)
+                     ": " (emms-track-name track))))))
+
+(defun emms-format-url-track-name (name)
+  "Format URL track name for better readability."
+  (url-unhex-string name))
 
 (defun emms-track-force-description (track)
   "Always return text that describes TRACK.
@@ -599,12 +738,35 @@ string), a confusing error message would result."
         desc
       (emms-track-simple-description track))))
 
+(defun emms-track-get-year (track)
+  "Get year of TRACK for display.
+There is the separation between the 'release date' and the
+'original date'.  This difference matters e.g. for
+re-releases (anniversaries and such) where the release date is
+more recent than the original release date.  In such cases the
+user probably wants the original release date so this is what we
+show."
+  (or
+   (emms-format-date-to-year (emms-track-get track 'info-date))
+   (emms-format-date-to-year (emms-track-get track 'info-originaldate))
+   (emms-track-get track 'info-year)
+   (emms-track-get track 'info-originalyear)))
+
+(defun emms-format-date-to-year (date)
+  "Try to extract year part from DATE.
+Return nil if the year cannot be extracted."
+  (when date
+    (let ((year (nth 5 (parse-time-string date))))
+      (if year (number-to-string year)
+        (when (string-match "^[ \t]*\\([0-9]\\{4\\}\\)" date)
+          (match-string 1 date))))))
+
 
 ;;; The Playlist
 
-;; Playlists are stored in buffers. The current playlist buffer is
-;; remembered in the `emms-playlist' variable. The buffer consists of
-;; any kind of data. Strings of text with a `emms-track' property are
+;; Playlists are stored in buffers.  The current playlist buffer is
+;; remembered in the `emms-playlist' variable.  The buffer consists of
+;; any kind of data.  Strings of text with a `emms-track' property are
 ;; the tracks in the buffer.
 
 (defvar emms-playlist-buffers nil
@@ -618,7 +780,7 @@ for that purpose.")
 (make-variable-buffer-local 'emms-playlist-selected-marker)
 
 (defvar emms-playlist-buffer-p nil
-  "Non-nil when the current buffer is an EMMS playlist.")
+  "Non-nil if the current buffer is an EMMS playlist.")
 (make-variable-buffer-local 'emms-playlist-buffer-p)
 
 (defun emms-playlist-ensure-playlist-buffer ()
@@ -632,32 +794,33 @@ for that purpose.")
    (list (let* ((buf-list (mapcar #'(lambda (buf)
                                       (list (buffer-name buf)))
                                   (emms-playlist-buffer-list)))
+                (sorted-buf-list (sort buf-list
+                                       #'(lambda (lbuf rbuf)
+                                           (< (length (car lbuf))
+                                              (length (car rbuf))))))
                 (default (or (and emms-playlist-buffer-p
                                   ;; default to current buffer
                                   (buffer-name))
                              ;; pick shortest buffer name, since it is
                              ;; likely to be a shared prefix
-                             (car (sort buf-list
-                                        #'(lambda (lbuf rbuf)
-                                            (< (length (car lbuf))
-                                               (length (car rbuf)))))))))
-           (completing-read "Playlist buffer to make current: "
-                            buf-list nil t default))))
+                             (car sorted-buf-list))))
+           (emms-completing-read "Playlist buffer to make current: "
+                                 sorted-buf-list nil t default))))
   (let ((buf (if buffer
                  (get-buffer buffer)
                (current-buffer))))
     (with-current-buffer buf
       (emms-playlist-ensure-playlist-buffer))
     (setq emms-playlist-buffer buf)
-    (when (interactive-p)
+    (when (called-interactively-p 'interactive)
       (message "Set current EMMS playlist buffer"))
     buf))
 
 (defun emms-playlist-new (&optional name)
   "Create a new playlist buffer.
-The buffer is named NAME, but made unique. NAME defaults to
-`emms-playlist-buffer-name'.
-If called interactively, the new buffer is also selected."
+The buffer is named NAME, but made unique.  NAME defaults to
+`emms-playlist-buffer-name'.  If called interactively, the new
+buffer is also selected."
   (interactive)
   (let ((buf (generate-new-buffer (or name
                                       emms-playlist-buffer-name))))
@@ -666,7 +829,7 @@ If called interactively, the new buffer is also selected."
         (funcall emms-playlist-default-major-mode))
       (setq emms-playlist-buffer-p t))
     (add-to-list 'emms-playlist-buffers buf)
-    (when (interactive-p)
+    (when (called-interactively-p 'interactive)
       (switch-to-buffer buf))
     buf))
 
@@ -721,8 +884,7 @@ If no current playlist exists, a new one is generated."
       (emms-playlist-clear))))
 
 (defun emms-playlist-clear ()
-  "Clear the current buffer.
-If no playlist exists, a new one is generated."
+  "Clear the current buffer."
   (interactive)
   (emms-playlist-ensure-playlist-buffer)
   (let ((inhibit-read-only t))
@@ -815,7 +977,7 @@ If no playlist exists, a new one is generated."
     (emms-playlist-selected-track)))
 
 (defun emms-playlist-selected-track-at-p (&optional point)
-  "Return non-nil when POINT (defaulting to point) is on the selected track."
+  "Return non-nil if POINT (defaulting to point) is on the selected track."
   (when emms-playlist-selected-marker
     (or (= emms-playlist-selected-marker
            (or point (point)))
@@ -1000,7 +1162,7 @@ This is supplying ARGS as arguments to the source."
 
 (defun emms-playlist-simple-update-track ()
   "Update the track at point.
-Since we don't do anything special with the track anyways, just
+Since we don't do anything special with the track anyway, just
 ignore this."
   nil)
 
@@ -1063,7 +1225,7 @@ ignore this."
 (defun emms-uniq-list (list stringify)
   "Compare stringfied element of list, and remove duplicate elements."
   ;; This uses a fast append list, keeping a pointer to the last cons
-  ;; cell of the list (TAIL). It might be worthwhile to provide an
+  ;; cell of the list (TAIL).  It might be worthwhile to provide an
   ;; abstraction for this eventually.
   (let* ((hash (make-hash-table :test 'equal))
          (result (cons nil nil))
@@ -1077,7 +1239,8 @@ ignore this."
     (cdr result)))
 
 (defun emms-playlist-simple-uniq ()
-  "Remove duplicate tracks"
+  "Remove duplicate tracks."
+  ;; TODO: This seems unnecessarily destructive.
   (emms-playlist-ensure-playlist-buffer)
   (widen)
   (let ((inhibit-read-only t)
@@ -1096,6 +1259,10 @@ ignore this."
           (emms-playlist-select pos)
         (emms-playlist-first)))))
 
+(defun emms-default-ok-track-function (track)
+  "A function which OKs all tracks for playing by default."
+  t)
+
 ;;; Helper functions
 (defun emms-property-region (pos prop)
   "Return a pair of the beginning and end of the property PROP at POS.
@@ -1134,18 +1301,18 @@ If POS does not contain PROP, try to find PROP just before POS."
 ;; It should use `emms-playlist-insert-track' to insert the tracks it
 ;; knows about.
 ;;
-;; The define-emms-source macro also defines functions emms-play-SOURCE
-;; and emms-add-SOURCE.  The former will replace the current playlist,
-;; while the latter will add to the end.
+;; The define-emms-source macro also defines functions
+;; emms-play-SOURCE and emms-add-SOURCE.  The former will replace the
+;; current playlist, while the latter will add to the end.
 
 (defmacro define-emms-source (name arglist &rest body)
   "Define a new EMMS source called NAME.
-This macro defines three functions: `emms-source-NAME', `emms-play-NAME'
-and `emms-add-NAME'. BODY should use `emms-playlist-insert-track'
-do insert all tracks to be played, which is exactly what
-`emms-source-NAME' will do.
-The other two functions will be simple wrappers around `emms-source-NAME';
-any `interactive' form that you specify in BODY will end up in these.
+This macro defines three functions: `emms-source-NAME',
+`emms-play-NAME' and `emms-add-NAME'.  BODY should use
+`emms-playlist-insert-track' to insert all tracks to be played,
+which is exactly what `emms-source-NAME' will do.  The other two
+functions will be simple wrappers around `emms-source-NAME'; any
+`interactive' form that you specify in BODY will end up in these.
 See emms-source-file.el for some examples."
   (let ((source-name (intern (format "emms-source-%s" name)))
         (source-play (intern (format "emms-play-%s" name)))
@@ -1237,18 +1404,19 @@ See emms-source-file.el for some examples."
 (defun emms-player (start stop playablep)
   "Create a new EMMS player.
 The start function will be START, and the stop function STOP.
-PLAYABLEP should return non-nil for tracks that this player can play.
+PLAYABLEP should return non-nil for tracks that this player can
+play.
 
-When trying to play a track, EMMS walks `emms-player-list'.
-For each player,it calls the PLAYABLEP function.
-The player corresponding to the first PLAYABLEP function that returns
-non-nil is used to play the track.
-To actually play the track, EMMS calls the START function,
-passing the chosen track as a parameter.
+When trying to play a track, EMMS walks through
+`emms-player-list'.  For each player, it calls the PLAYABLEP
+function.  The player corresponding to the first PLAYABLEP
+function that returns non-nil is used to play the track.  To
+actually play the track, EMMS calls the START function, passing
+the chosen track as a parameter.
 
-If the user tells EMMS to stop playing, the STOP function is called.
-Once the player has finished playing, it should call `emms-player-stopped'
-to let EMMS know."
+If the user tells EMMS to stop playing, the STOP function is
+called.  Once the player has finished playing, it should call
+`emms-player-stopped' to let EMMS know."
   (let ((p (emms-dictionary '*player*)))
     (emms-player-set p 'start start)
     (emms-player-set p 'stop stop)
@@ -1271,8 +1439,8 @@ to let EMMS know."
 
 (defun emms-player-for (track)
   "Return an EMMS player capable of playing TRACK.
-This will be the first player whose PLAYABLEP function returns non-nil,
-or nil if no such player exists."
+This will be the first player whose PLAYABLEP function returns
+non-nil, or nil if no such player exists."
   (let ((lis emms-player-list))
     (while (and lis
                 (not (funcall (emms-player-get (car lis) 'playablep)
@@ -1299,7 +1467,7 @@ or nil if no such player exists."
   "Declare that the given EMMS PLAYER has started.
 This should only be done by the current player itself."
   (setq emms-player-playing-p player
-        emms-player-paused-p nil)
+        emms-player-paused-p  nil)
   (run-hooks 'emms-player-started-hook))
 
 (defun emms-player-stop ()
@@ -1346,8 +1514,8 @@ This should only be done by the current player itself."
 
 (defun emms-player-seek (seconds)
   "Seek the current player by SECONDS seconds.
-This can be a floating point number for fractions of a second,
-or negative to seek backwards."
+This can be a floating point number for fractions of a second, or
+negative to seek backwards."
   (if (not emms-player-playing-p)
       (error "Can't seek player, nothing playing right now")
     (let ((seek (emms-player-get emms-player-playing-p 'seek)))
@@ -1358,8 +1526,8 @@ or negative to seek backwards."
 
 (defun emms-player-seek-to (seconds)
   "Seek the current player to SECONDS seconds.
-This can be a floating point number for fractions of a second,
-or negative to seek backwards."
+This can be a floating point number for fractions of a second, or
+negative to seek backwards."
   (if (not emms-player-playing-p)
       (error "Can't seek-to player, nothing playing right now")
     (let ((seek (emms-player-get emms-player-playing-p 'seek-to)))
